import json
import logging
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests
from requests.exceptions import RequestException

from checkout.exceptions.http_exceptions import ClientErrorException


class HttpClient:
    def __init__(
        self,
        base_url: str,
        timeout: Optional[int] = None,
        logger: Optional[logging.Logger] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Initialize the HTTP client.

        :param base_url: Base URL for the API.
        :param timeout: Timeout for requests in seconds.
        :param logger: Logger instance for logging requests and responses.
        """
        self.base_url = self._sanitize_base_url(base_url)
        self.timeout = timeout or 10
        self.logger = logger or self._default_logger()
        self.headers = headers or {"Content-Type": "application/json"}

    @staticmethod
    def _sanitize_base_url(base_url: str) -> str:
        """Ensure the base URL does not end with a trailing slash."""
        return urljoin(str(base_url).rstrip("/") + "/", "")

    def post(self, endpoint: str, json: Dict[str, Any], headers: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
        """
        Make an HTTP POST request to the specified endpoint.

        :param endpoint: The API endpoint.
        :param json: The data to include in the request body.
        :param headers: Optional HTTP headers.
        :return: Parsed JSON response as a dictionary.
        :raises ClientErrorException: For client-side HTTP errors (4xx).
        :raises ServerErrorException: For server-side HTTP errors (5xx).
        """
        url = self._construct_url(endpoint)
        try:
            self._log_request(url, json)
            final_headers = {**self.headers, **(headers or {})}
            response = requests.post(url=url, json=json, headers=final_headers, timeout=self.timeout)
            response.raise_for_status()

            self._log_response(response)
            return response.json()
        except RequestException:
            raise ClientErrorException.from_response(response)
        except Exception as e:
            self._handle_generic_exception(e)
            raise

    def _construct_url(self, endpoint: str) -> str:
        """Construct the full URL for the API endpoint."""
        return f"{urljoin(str(self.base_url).rstrip('/') + '/', '')}{endpoint.lstrip('/')}"

    def _log_request(self, url: str, payload: Dict[str, Any]) -> None:
        """Log the details of an outgoing request."""
        try:
            payload_str = json.dumps(payload, ensure_ascii=False)
        except (TypeError, ValueError) as e:
            payload_str = f"Error serializing payload: {e}"

        self.logger.debug(f"REQUEST to {url}: {payload_str}")

    def _log_response(self, response: requests.Response) -> None:
        """Log the details of a received response."""
        self.logger.debug("RESPONSE", {"status_code": response.status_code, "result": response.text})

    def _handle_generic_exception(self, exception: Exception) -> None:
        """Handle unexpected exceptions."""
        self.logger.warning("GENERIC_EXCEPTION", {"error": str(exception)})
        raise exception

    @staticmethod
    def _default_logger() -> logging.Logger:
        """
        Create and configure a default logger.
        """
        logger = logging.getLogger("HttpClient")
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger



from typing import Any, Dict, List, Union

from checkout.entities.payment import Payment


class DispersionPayment(Payment):
    dispersion: List[Payment] = []

    def __init__(self, **data: Dict[str, Any]):
        """
        Initialize DispersionPayment object and process dispersion payments.
        """
        payment_fields = self._extract_payment_fields(data)
        super().__init__(**payment_fields)
        if "dispersion" in data:
            self.set_dispersion(data["dispersion"])

    def _extract_payment_fields(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Extract and validate fields for initializing the base Payment class.
        """
        return {
            "reference": data.get("reference", ""),
            "description": data.get("description", ""),
            "amount": data.get("amount"),
            "allowPartial": data.get("allowPartial", False),
            "shipping": data.get("shipping"),
            "items": data.get("items", []),
            "recurring": data.get("recurring"),
            "payment": data.get("payment"),
            "discount": data.get("discount"),
            "subscribe": data.get("subscribe", False),
            "agreement": data.get("agreement"),
            "agreementType": data.get("agreementType", ""),
            "modifiers": data.get("modifiers", []),
        }

    def set_dispersion(self, data: Union[List[Dict], Dict]) -> "DispersionPayment":
        """
        Set the dispersion payments.
        """
        self.dispersion = []
        for payment_data in data:
            payment = Payment(**payment_data) if isinstance(payment_data, dict) else payment_data
            self.dispersion.append(payment)
        return self

    def dispersion_to_dict(self) -> List[Dict]:
        """
        Convert the list of dispersion payments to a list of dictionaries.
        """
        return [payment.to_dict() for payment in self.dispersion]

    def to_dict(self) -> Dict:
        """
        Convert the DispersionPayment object to a dictionary.
        """
        base_data = super().to_dict()
        base_data["dispersion"] = self.dispersion_to_dict()
        return base_data



import unittest

from checkout.entities.amount import Amount
from checkout.entities.amount_detail import AmountDetail
from checkout.entities.tax_detail import TaxDetail


class AmountTest(unittest.TestCase):

    def test_amount_initialization(self):
        """
        Test initializing the Amount class with default values.
        """
        amount = Amount(total=100.0, currency="COP")

        assert amount.total == 100.0
        assert amount.currency == "COP"
        assert amount.taxes == []
        assert amount.details == []
        assert amount.tip is None
        assert amount.insurance is None

    def test_set_taxes(self):
        """
        Test setting taxes using a mix of dictionaries and `TaxDetail` instances.
        """
        tax_data = [
            {"kind": "VAT", "amount": 10.0},
            TaxDetail(kind="SERVICE", amount=5.0),
        ]

        amount = Amount(total=100.0, currency="COP")
        amount.set_taxes(tax_data)

        assert len(amount.taxes) == 2
        assert amount.taxes[0].kind == "VAT"
        assert amount.taxes[0].amount == 10.0
        assert amount.taxes[1].kind == "SERVICE"
        assert amount.taxes[1].amount == 5.0

    def test_set_details(self):
        """
        Test setting details using a mix of dictionaries and `AmountDetail` instances.
        """
        details_data = [
            {"kind": "FEE", "amount": 3.0},
            AmountDetail(kind="DISCOUNT", amount=2.0),
        ]

        amount = Amount(total=100.0, currency="COP")
        amount.set_details(details_data)

        assert len(amount.details) == 2
        assert amount.details[0].kind == "FEE"
        assert amount.details[0].amount == 3.0
        assert amount.details[1].kind == "DISCOUNT"
        assert amount.details[1].amount == 2.0

    def test_taxes_to_dict(self):
        """
        Test converting taxes to an array of dictionaries.
        """
        tax_data = [
            {"kind": "VAT", "amount": 10.0},
            TaxDetail(kind="SERVICE", amount=5.0),
        ]

        amount = Amount(total=100.0, currency="COP")
        amount.set_taxes(tax_data)

        taxes_dict = amount.taxes_to_dict()
        assert len(taxes_dict) == 2
        assert taxes_dict[0] == {"amount": 10.0, "base": None, "kind": "VAT"}
        assert taxes_dict[1] == {"amount": 5.0, "base": None, "kind": "SERVICE"}

    def test_details_to_dict(self):
        """
        Test converting details to an array of dictionaries.
        """
        details_data = [
            {"kind": "FEE", "amount": 3.0},
            AmountDetail(kind="DISCOUNT", amount=2.0),
        ]

        amount = Amount(total=100.0, currency="COP")
        amount.set_details(details_data)

        details_array = amount.details_to_dict()
        assert len(details_array) == 2
        assert details_array[0] == {"kind": "FEE", "amount": 3.0}
        assert details_array[1] == {"kind": "DISCOUNT", "amount": 2.0}

    def test_to_dict(self):
        """
        Test converting the Amount object to a dictionary including taxes and details.
        """
        tax_data = [
            {"kind": "VAT", "amount": 10.0},
            TaxDetail(kind="SERVICE", amount=5.0),
        ]
        details_data = [
            {"kind": "FEE", "amount": 3.0},
            AmountDetail(kind="DISCOUNT", amount=2.0),
        ]

        amount = Amount(total=100.0, currency="COP", tip=15.0, insurance=20.0)
        amount.set_taxes(tax_data)
        amount.set_details(details_data)

        amount_dict = amount.to_dict()
        expected_dict = {
            "currency": "COP",
            "total": 100.0,
            "taxes": [{"kind": "VAT", "amount": 10.0, "base": None}, {"kind": "SERVICE", "amount": 5.0, "base": None}],
            "details": [{"kind": "FEE", "amount": 3.0}, {"kind": "DISCOUNT", "amount": 2.0}],
            "tip": 15.0,
            "insurance": 20.0,
        }

        assert amount_dict == expected_dict
